/*
   sshpubkey.c 
   
   Author: Vesa Suontama <vsuontam@ssh.fi>
   
   Copyright (c) 2001 SSH Communications Security Corp, Helsinki, Finland
   All rights reserved
   
   Created Tue May 08 11:59:24 2001. 
   
   */

#include "sshincludes.h"
#include "sshpubkey.h"

#define SSH_DEBUG_MODULE "SshPKB"

/* Get type of the public key from the keyblob given as `data' string
   whose length is `len' bytes. The call detects type of the public
   key and fills corresponding value into `kind'. If comment is not
   NULL, and the key format supports visible comments, a copy of
   comment string is returned in `comment'.

   The function returns SSH_CRYPTO_OK if the key format was
   determined, and SSH_CRYPTO_UNKNOWN_KEY_TYPE if not. */
SshCryptoStatus
ssh_pkb_get_info(const unsigned char *data, size_t len,
                 SshPKBType *kind, char **comment)
{



  SshPublicKey pk;

  if (comment)
    *comment = NULL;

  /* Check for SSH public key. */
  if (ssh_public_key_import(data, len, &pk) == SSH_CRYPTO_OK)
    {
      ssh_public_key_free(pk);
      *kind = SSH_PKB_SSH;
      return SSH_CRYPTO_OK;
    }












  SSH_DEBUG(SSH_D_MIDOK, ("Could not deduce the public key type"));
  *kind = SSH_PKB_UNKNOWN;
  return SSH_CRYPTO_UNKNOWN_KEY_TYPE;
}


/* Returns information about the key type. *Needs_secret is set to
   TRUE if decoding if the blob requires some secret code.  A
   printable name of the type is returned in
   key_type_name_ret. Returns TRUE on success and FALSE otherwise. */
Boolean
ssh_pkb_get_type_info(SshPKBType type,
                      Boolean *needs_secret,
                      const char **key_type_name_ret)
{
  /* XXX pretty basic with so few key types. :) */
  /* Table of key types and their 'properties' */
  static const struct SshPKBPropertiesRec
    {
      SshPKBType type;
      Boolean needs_secret;
      const char *name;
    } ssh_pkb_properties[] =
      {
        { SSH_PKB_UNKNOWN, FALSE, "Unkown"},
        { SSH_PKB_SSH, FALSE, "SSH Key"},



      };
  int i, l;
  /* Find the right type. */
  l = sizeof(ssh_pkb_properties) / sizeof(struct SshPKBPropertiesRec);
  for (i = 0; i < l; i++)
    {
      if (type == ssh_pkb_properties[i].type)
        {
          /* Type found. */
          if (needs_secret)
            *needs_secret = ssh_pkb_properties[i].needs_secret;
          if (key_type_name_ret)
            *key_type_name_ret = ssh_pkb_properties[i].name;
          return TRUE;
        }
    }
  /* Type was not found. */
  return FALSE;  
}
































/* Get public key out from the data blob given as `data' and `len'.
   If the `kind' argument specifies an encrypted key format, the
   password (or passphares) needed to decrypt the key is given as
   `password' and `password_len'.

   The function returns SSH_CRYPTO_OK if the public key was
   successfully decoded from the data. If the password is incorrect,
   the return value shall be SSH_CRYPTO_INVALID_PASSPHRASE, and if the
   key blob is not of specified type, the return value shall be
   SSH_CRYPTO_CORRUPTED_KEY_FORMAT.
*/
SshCryptoStatus
ssh_pkb_decode(SshPKBType kind,
               const unsigned char *data, size_t len,
               const unsigned char *password, size_t password_len,
               SshPublicKey *key)
{
  switch (kind)
    {
    case SSH_PKB_SSH:
      if (ssh_public_key_import(data, len, key) == SSH_CRYPTO_OK)
        return SSH_CRYPTO_OK;
      break;
      








    case SSH_PKB_UNKNOWN: 
    default: SSH_DEBUG(SSH_D_FAIL, ("pkb decode with unknown key type"));
      break;
    }
  return SSH_CRYPTO_CORRUPTED_KEY_FORMAT;
}




